@@ -17,4 +17,15 @@ pub struct LocationProps {
     pub logout_href: String,
 }
 
-pub type LocationRedirect = Redirect<LocationRedirector>;
+pub mod oauth2 {
+    use super::*;
+    use crate::agent::client::OAuth2Client as Client;
+    pub type LocationRedirect = Redirect<Client, LocationRedirector>;
+}
+
+#[cfg(feature = "openid")]
+pub mod openid {
+    use super::*;
+    use crate::agent::client::OpenIdClient as Client;
+    pub type LocationRedirect = Redirect<Client, LocationRedirector>;
+}
